@@ -13,6 +13,7 @@ use crate::system_info::SystemInfo;
 use crate::{FrameSymbolizer, SymbolStats};
 use minidump::system_info::Cpu;
 use minidump::*;
+use minidump_common::utils::basename;
 use serde_json::json;
 
 /// Indicates how well the instruction pointer derived during
@@ -326,13 +327,6 @@ impl FrameSymbolizer for StackFrame {
     }
 }
 
-fn basename(f: &str) -> &str {
-    match f.rfind(|c| c == '/' || c == '\\') {
-        None => f,
-        Some(index) => &f[(index + 1)..],
-    }
-}
-
 fn print_registers<T: Write>(f: &mut T, ctx: &MinidumpContext) -> io::Result<()> {
     let registers: Cow<HashSet<&str>> = match ctx.valid {
         MinidumpContextValidity::All => {
