@@ -3,8 +3,9 @@
 
 use crate::process_state::*;
 use crate::stackwalker::walk_stack;
-use crate::{string_symbol_supplier, Symbolizer};
+use crate::{string_symbol_supplier, Symbolizer, SystemInfo};
 use minidump::format::CONTEXT_ARM;
+use minidump::system_info::{Cpu, Os};
 use minidump::*;
 use std::collections::HashMap;
 use test_assembler::*;
@@ -12,6 +13,7 @@ use test_assembler::*;
 struct TestFixture {
     pub raw: CONTEXT_ARM,
     pub modules: MinidumpModuleList,
+    pub system_info: SystemInfo,
     pub symbols: HashMap<String, String>,
 }
 
@@ -25,11 +27,20 @@ impl TestFixture {
                 MinidumpModule::new(0x40000000, 0x10000, "module1"),
                 MinidumpModule::new(0x50000000, 0x10000, "module2"),
             ]),
+            system_info: SystemInfo {
+                os: Os::Ios,
+                os_version: None,
+                os_build: None,
+                cpu: Cpu::Arm,
+                cpu_info: None,
+                cpu_microcode_version: None,
+                cpu_count: 1,
+            },
             symbols: HashMap::new(),
         }
     }
 
-    pub fn walk_stack(&self, stack: Section) -> CallStack {
+    pub async fn walk_stack(&self, stack: Section) -> CallStack {
         let context = MinidumpContext {
             raw: MinidumpRawContext::Arm(self.raw.clone()),
             valid: MinidumpContextValidity::All,
@@ -48,8 +59,10 @@ impl TestFixture {
             &Some(&context),
             Some(&stack_memory),
             &self.modules,
+            &self.system_info,
             &symbolizer,
         )
+        .await
     }
 
     pub fn add_symbols(&mut self, name: String, symbols: String) {
@@ -57,8 +70,8 @@ impl TestFixture {
     }
 }
 
-#[test]
-fn test_simple() {
+#[tokio::test]
+async fn test_simple() {
     let mut f = TestFixture::new();
     let stack = Section::new();
     stack.start().set_const(0x80000000);
@@ -69,15 +82,15 @@ fn test_simple() {
     f.raw.set_register("pc", 0x4000c020);
     f.raw.set_register("fp", 0x80000000);
 
-    let s = f.walk_stack(stack);
+    let s = f.walk_stack(stack).await;
     assert_eq!(s.frames.len(), 1);
     let f = &s.frames[0];
     let m = f.module.as_ref().unwrap();
     assert_eq!(m.code_file(), "module1");
 }
 
-#[test]
-fn test_scan_without_symbols() {
+#[tokio::test]
+async fn test_scan_without_symbols() {
     // Scanning should work without any symbols
     let mut f = TestFixture::new();
     let mut stack = Section::new();
@@ -105,10 +118,13 @@ fn test_scan_without_symbols() {
         .append_repeated(0, 32); // end of stack
 
     f.raw.set_register("pc", 0x40005510);
+    // set an invalid non-zero value for the frame pointer
+    // to force stack scanning
+    f.raw.set_register("fp", 0x00000001);
     f.raw
         .set_register("sp", stack.start().value().unwrap() as u32);
 
-    let s = f.walk_stack(stack);
+    let s = f.walk_stack(stack).await;
     assert_eq!(s.frames.len(), 3);
 
     {
@@ -163,8 +179,8 @@ fn test_scan_without_symbols() {
     }
 }
 
-#[test]
-fn test_scan_first_frame() {
+#[tokio::test]
+async fn test_scan_first_frame() {
     // The first (context) frame gets extra long scans, this test checks that.
     let mut f = TestFixture::new();
     let mut stack = Section::new();
@@ -194,10 +210,13 @@ fn test_scan_first_frame() {
         .append_repeated(0, 64); // end of stack
 
     f.raw.set_register("pc", 0x40005510);
+    // set an invalid non-zero value for the frame pointer
+    // to force stack scanning
+    f.raw.set_register("fp", 0x00000001);
     f.raw
         .set_register("sp", stack.start().value().unwrap() as u32);
 
-    let s = f.walk_stack(stack);
+    let s = f.walk_stack(stack).await;
     assert_eq!(s.frames.len(), 2);
 
     {
@@ -230,8 +249,104 @@ fn test_scan_first_frame() {
     }
 }
 
-#[test]
-fn test_frame_pointer() {
+#[tokio::test]
+async fn test_invalid_lr() {
+    let mut f = TestFixture::new();
+    f.system_info.os = Os::Linux;
+
+    let mut stack = Section::new();
+    stack.start().set_const(0x80000000);
+
+    let lr = Label::new();
+    let return_address1 = 0x50000100u32;
+    let return_address2 = 0x50000900u32;
+    let frame1_sp = Label::new();
+    let frame2_sp = Label::new();
+    let frame1_fp = Label::new();
+    let frame2_fp = Label::new();
+
+    stack = stack
+        // frame 0
+        .append_repeated(0, 32) // space
+        .mark(&lr) // the LR points to something on the stack
+        .D32(0x0000000D) // junk that's not
+        .D32(0xF0000000) // a return address
+        .mark(&frame1_fp) // next fp will point to the next value
+        .D32(&frame2_fp) // save current frame pointer
+        .D32(return_address1) // save current link register
+        .mark(&frame1_sp)
+        // frame 1
+        .append_repeated(0, 32) // space
+        .D32(0x0000000D) // junk that's not
+        .D32(0xF0000000) // a return address
+        .mark(&frame2_fp)
+        .D32(0)
+        .D32(return_address2)
+        .mark(&frame2_sp);
+
+    f.raw.set_register("pc", 0x40005510);
+    f.raw.set_register("lr", lr.value().unwrap() as u32);
+    f.raw.set_register("fp", frame1_fp.value().unwrap() as u32);
+    f.raw
+        .set_register("sp", stack.start().value().unwrap() as u32);
+
+    let s = f.walk_stack(stack).await;
+    assert_eq!(s.frames.len(), 3);
+
+    {
+        // Frame 0
+        let frame = &s.frames[0];
+        assert_eq!(frame.trust, FrameTrust::Context);
+        assert_eq!(frame.context.valid, MinidumpContextValidity::All);
+    }
+
+    {
+        // Frame 1
+        let frame = &s.frames[1];
+        let valid = &frame.context.valid;
+        assert_eq!(frame.trust, FrameTrust::Scan);
+        if let MinidumpContextValidity::Some(ref which) = valid {
+            assert_eq!(which.len(), 2);
+        } else {
+            unreachable!();
+        }
+
+        if let MinidumpRawContext::Arm(ctx) = &frame.context.raw {
+            assert_eq!(ctx.get_register("pc", valid).unwrap(), return_address1);
+            assert_eq!(
+                ctx.get_register("sp", valid).unwrap(),
+                frame1_sp.value().unwrap() as u32
+            );
+        } else {
+            unreachable!();
+        }
+    }
+
+    {
+        // Frame 2
+        let frame = &s.frames[2];
+        let valid = &frame.context.valid;
+        assert_eq!(frame.trust, FrameTrust::Scan);
+        if let MinidumpContextValidity::Some(ref which) = valid {
+            assert_eq!(which.len(), 2);
+        } else {
+            unreachable!();
+        }
+
+        if let MinidumpRawContext::Arm(ctx) = &frame.context.raw {
+            assert_eq!(ctx.get_register("pc", valid).unwrap(), return_address2);
+            assert_eq!(
+                ctx.get_register("sp", valid).unwrap(),
+                frame2_sp.value().unwrap() as u32
+            );
+        } else {
+            unreachable!();
+        }
+    }
+}
+
+#[tokio::test]
+async fn test_frame_pointer() {
     // Frame-pointer-based unwinding
     let mut f = TestFixture::new();
     let mut stack = Section::new();
@@ -272,7 +387,7 @@ fn test_frame_pointer() {
     f.raw
         .set_register("sp", stack.start().value().unwrap() as u32);
 
-    let s = f.walk_stack(stack);
+    let s = f.walk_stack(stack).await;
     assert_eq!(s.frames.len(), 3);
 
     {
@@ -334,8 +449,8 @@ fn test_frame_pointer() {
     }
 }
 
-#[test]
-fn test_frame_pointer_infinite_equality() {
+#[tokio::test]
+async fn test_frame_pointer_infinite_equality() {
     // Leaf functions on Arm are allowed to not update the stack pointer, so
     // it's valid for the frame pointer analysis to conclude that the stack
     // pointer doesn't change. However we must only provide this allowance
@@ -386,7 +501,7 @@ fn test_frame_pointer_infinite_equality() {
     f.raw
         .set_register("sp", stack.start().value().unwrap() as u32);
 
-    let s = f.walk_stack(stack);
+    let s = f.walk_stack(stack).await;
     assert_eq!(s.frames.len(), 2);
 
     {
@@ -426,7 +541,7 @@ fn test_frame_pointer_infinite_equality() {
     // Never get to frame 2, alas!
 }
 
-const CALLEE_SAVE_REGS: &[&str] = &["pc", "sp", "r4", "r5", "r6", "r7", "r8", "r9", "r10", "r11"];
+const CALLEE_SAVE_REGS: &[&str] = &["pc", "sp", "r4", "r5", "r6", "r7", "r8", "r9", "r10", "fp"];
 
 fn init_cfi_state() -> (TestFixture, Section, CONTEXT_ARM, MinidumpContextValidity) {
     let mut f = TestFixture::new();
@@ -492,13 +607,13 @@ fn init_cfi_state() -> (TestFixture, Section, CONTEXT_ARM, MinidumpContextValidi
     (f, stack, expected, expected_valid)
 }
 
-fn check_cfi(
+async fn check_cfi(
     f: TestFixture,
     stack: Section,
     expected: CONTEXT_ARM,
     expected_valid: MinidumpContextValidity,
 ) {
-    let s = f.walk_stack(stack);
+    let s = f.walk_stack(stack).await;
     assert_eq!(s.frames.len(), 2);
 
     {
@@ -536,8 +651,8 @@ fn check_cfi(
     unreachable!();
 }
 
-#[test]
-fn test_cfi_at_4000() {
+#[tokio::test]
+async fn test_cfi_at_4000() {
     let (mut f, mut stack, expected, expected_valid) = init_cfi_state();
 
     stack = stack.append_repeated(0, 120);
@@ -545,11 +660,11 @@ fn test_cfi_at_4000() {
     f.raw.set_register("pc", 0x40004000);
     f.raw.set_register("lr", 0x40005510);
 
-    check_cfi(f, stack, expected, expected_valid);
+    check_cfi(f, stack, expected, expected_valid).await;
 }
 
-#[test]
-fn test_cfi_at_4001() {
+#[tokio::test]
+async fn test_cfi_at_4001() {
     let (mut f, mut stack, mut expected, expected_valid) = init_cfi_state();
 
     let frame1_sp = Label::new();
@@ -565,11 +680,11 @@ fn test_cfi_at_4001() {
     f.raw.iregs[4] = 0x635adc9f;
     f.raw.set_register("fp", 0xbe145fc4);
 
-    check_cfi(f, stack, expected, expected_valid);
+    check_cfi(f, stack, expected, expected_valid).await;
 }
 
-#[test]
-fn test_cfi_at_4002() {
+#[tokio::test]
+async fn test_cfi_at_4002() {
     let (mut f, mut stack, mut expected, expected_valid) = init_cfi_state();
 
     let frame1_sp = Label::new();
@@ -592,11 +707,11 @@ fn test_cfi_at_4002() {
     f.raw.iregs[7] = 0xa449829f; // distinct callee r7
     f.raw.set_register("fp", 0xbe145fc4);
 
-    check_cfi(f, stack, expected, expected_valid);
+    check_cfi(f, stack, expected, expected_valid).await;
 }
 
-#[test]
-fn test_cfi_at_4003() {
+#[tokio::test]
+async fn test_cfi_at_4003() {
     let (mut f, mut stack, mut expected, mut expected_valid) = init_cfi_state();
 
     let frame1_sp = Label::new();
@@ -619,11 +734,11 @@ fn test_cfi_at_4003() {
     f.raw.set_register("pc", 0x40004003);
     f.raw.iregs[1] = 0xfb756319;
 
-    check_cfi(f, stack, expected, expected_valid);
+    check_cfi(f, stack, expected, expected_valid).await;
 }
 
-#[test]
-fn test_cfi_at_4004() {
+#[tokio::test]
+async fn test_cfi_at_4004() {
     // Should be the same as 4003
     let (mut f, mut stack, mut expected, mut expected_valid) = init_cfi_state();
 
@@ -647,11 +762,11 @@ fn test_cfi_at_4004() {
     f.raw.set_register("pc", 0x40004004);
     f.raw.iregs[1] = 0xfb756319;
 
-    check_cfi(f, stack, expected, expected_valid);
+    check_cfi(f, stack, expected, expected_valid).await;
 }
 
-#[test]
-fn test_cfi_at_4005() {
+#[tokio::test]
+async fn test_cfi_at_4005() {
     let (mut f, mut stack, mut expected, mut expected_valid) = init_cfi_state();
 
     let frame1_sp = Label::new();
@@ -674,11 +789,11 @@ fn test_cfi_at_4005() {
     f.raw.set_register("pc", 0x40004005);
     f.raw.iregs[1] = 0xfb756319;
 
-    check_cfi(f, stack, expected, expected_valid);
+    check_cfi(f, stack, expected, expected_valid).await;
 }
 
-#[test]
-fn test_cfi_at_4006() {
+#[tokio::test]
+async fn test_cfi_at_4006() {
     // Here we provide an explicit rule for the PC, and have the saved .ra be
     // bogus.
 
@@ -705,11 +820,11 @@ fn test_cfi_at_4006() {
     f.raw.set_register("pc", 0x40004006);
     f.raw.iregs[1] = 0xfb756319;
 
-    check_cfi(f, stack, expected, expected_valid);
+    check_cfi(f, stack, expected, expected_valid).await;
 }
 
-#[test]
-fn test_cfi_reject_backwards() {
+#[tokio::test]
+async fn test_cfi_reject_backwards() {
     // Check that we reject rules that would cause the stack pointer to
     // move in the wrong direction.
 
@@ -721,12 +836,12 @@ fn test_cfi_reject_backwards() {
     f.raw.set_register("sp", 0x80000000);
     f.raw.set_register("lr", 0x40005510);
 
-    let s = f.walk_stack(stack);
+    let s = f.walk_stack(stack).await;
     assert_eq!(s.frames.len(), 1);
 }
 
-#[test]
-fn test_cfi_reject_bad_exprs() {
+#[tokio::test]
+async fn test_cfi_reject_bad_exprs() {
     // Check that we reject rules whose expressions' evaluation fails.
 
     let (mut f, mut stack, _expected, _expected_valid) = init_cfi_state();
@@ -736,6 +851,161 @@ fn test_cfi_reject_bad_exprs() {
     f.raw.set_register("pc", 0x40007000);
     f.raw.set_register("sp", 0x80000000);
 
-    let s = f.walk_stack(stack);
+    let s = f.walk_stack(stack).await;
     assert_eq!(s.frames.len(), 1);
 }
+
+#[tokio::test]
+async fn test_frame_pointer_overflow() {
+    // Make sure we don't explode when trying frame pointer analysis on a value
+    // that will overflow.
+
+    type Pointer = u32;
+    let stack_max: Pointer = Pointer::MAX;
+    let stack_size: Pointer = 1000;
+    let bad_frame_ptr: Pointer = stack_max;
+
+    let mut f = TestFixture::new();
+    let mut stack = Section::new();
+    let stack_start: Pointer = stack_max - stack_size;
+    stack.start().set_const(stack_start as u64);
+
+    stack = stack
+        // frame 0
+        .append_repeated(0, stack_size as usize); // junk, not important to the test
+
+    f.raw.set_register("pc", 0x7a100000);
+    f.raw.set_register("fp", bad_frame_ptr);
+    f.raw
+        .set_register("sp", stack.start().value().unwrap() as Pointer);
+    f.raw.set_register("lr", 0x7b302000);
+
+    let s = f.walk_stack(stack).await;
+    assert_eq!(s.frames.len(), 1);
+
+    // As long as we don't panic, we're good!
+}
+
+#[tokio::test]
+async fn test_frame_pointer_overflow_nonsense_32bit_stack() {
+    // same as test_frame_pointer_overflow, but we're going to abuse the fact
+    // that rust-minidump prefers representing things in 64-bit to create
+    // impossible stack addresses that overflow 32-bit integers but appear
+    // valid in 64-bit. By doing this memory reads will "succeed" but
+    // pointer math done in the native pointer width will overflow and
+    // everything will be sad.
+
+    type Pointer = u32;
+    let pointer_size: u64 = std::mem::size_of::<Pointer>() as u64;
+    let stack_max: u64 = Pointer::MAX as u64 + pointer_size * 2;
+    let stack_size: u64 = 1000;
+    let bad_frame_ptr: u64 = Pointer::MAX as u64 - pointer_size;
+
+    let mut f = TestFixture::new();
+    let mut stack = Section::new();
+    let stack_start: u64 = stack_max - stack_size;
+    stack.start().set_const(stack_start as u64);
+
+    stack = stack
+        // frame 0
+        .append_repeated(0, 1000); // junk, not important to the test
+
+    f.raw.set_register("pc", 0x7a100000);
+    f.raw.set_register("fp", bad_frame_ptr as u32);
+    f.raw
+        .set_register("sp", stack.start().value().unwrap() as Pointer);
+    f.raw.set_register("lr", 0x7b302000);
+
+    let s = f.walk_stack(stack).await;
+    assert_eq!(s.frames.len(), 1);
+
+    // As long as we don't panic, we're good!
+}
+
+#[tokio::test]
+async fn test_frame_pointer_barely_no_overflow() {
+    // This is a simple frame pointer test but with the all the values pushed
+    // as close to the upper memory boundary as possible, to confirm that
+    // our code doesn't randomly overflow *AND* isn't overzealous in
+    // its overflow guards.
+
+    let mut f = TestFixture::new();
+    let mut stack = Section::new();
+
+    type Pointer = u32;
+    let stack_max: Pointer = Pointer::MAX;
+    let pointer_size: Pointer = std::mem::size_of::<Pointer>() as Pointer;
+    let stack_size: Pointer = pointer_size * 3;
+
+    let stack_start: Pointer = stack_max - stack_size;
+    let return_address: Pointer = 0x7b302000;
+    stack.start().set_const(stack_start as u64);
+
+    let frame0_fp = Label::new();
+    let frame1_sp = Label::new();
+    let frame1_fp = Label::new();
+
+    stack = stack
+        // frame 0
+        .mark(&frame0_fp)
+        .D32(&frame1_fp) // caller-pushed %rbp
+        .D32(return_address) // actual return address
+        // frame 1
+        .mark(&frame1_sp)
+        .mark(&frame1_fp) // end of stack
+        .D32(0);
+
+    f.raw.set_register("pc", 0x7a100000);
+    f.raw
+        .set_register("fp", frame0_fp.value().unwrap() as Pointer);
+    f.raw
+        .set_register("sp", stack.start().value().unwrap() as Pointer);
+    f.raw.set_register("lr", return_address);
+
+    let s = f.walk_stack(stack).await;
+    assert_eq!(s.frames.len(), 2);
+
+    {
+        // Frame 0
+        let frame = &s.frames[0];
+        let valid = &frame.context.valid;
+        assert_eq!(frame.trust, FrameTrust::Context);
+        assert_eq!(frame.context.valid, MinidumpContextValidity::All);
+
+        if let MinidumpRawContext::Arm(ctx) = &frame.context.raw {
+            assert_eq!(
+                ctx.get_register("fp", valid).unwrap(),
+                frame0_fp.value().unwrap() as Pointer
+            );
+        } else {
+            unreachable!();
+        }
+    }
+
+    {
+        // Frame 1
+        let frame = &s.frames[1];
+        let valid = &frame.context.valid;
+        assert_eq!(frame.trust, FrameTrust::FramePointer);
+        if let MinidumpContextValidity::Some(ref which) = valid {
+            assert_eq!(which.len(), 4);
+        } else {
+            unreachable!();
+        }
+
+        if let MinidumpRawContext::Arm(ctx) = &frame.context.raw {
+            assert_eq!(ctx.get_register("pc", valid).unwrap(), return_address);
+            assert_eq!(
+                ctx.get_register("sp", valid).unwrap(),
+                frame1_sp.value().unwrap() as Pointer
+            );
+            assert_eq!(
+                ctx.get_register("fp", valid).unwrap(),
+                frame1_fp.value().unwrap() as Pointer
+            );
+            assert_eq!(ctx.get_register("lr", valid).unwrap(), return_address);
+        } else {
+            unreachable!();
+        }
+    }
+}
